@@ -26,6 +26,8 @@ from dnf.module.exceptions import EnableMultipleStreamsException
 from dnf.util import logger
 from dnf.i18n import _, P_, ucd
 
+import functools
+
 STATE_DEFAULT = libdnf.module.ModulePackageContainer.ModuleState_DEFAULT
 STATE_ENABLED = libdnf.module.ModulePackageContainer.ModuleState_ENABLED
 STATE_DISABLED = libdnf.module.ModulePackageContainer.ModuleState_DISABLED
@@ -33,6 +35,11 @@ STATE_UNKNOWN = libdnf.module.ModulePackageContainer.ModuleState_UNKNOWN
 MODULE_TABLE_HINT = _("\n\nHint: [d]efault, [e]nabled, [x]disabled, [i]nstalled")
 MODULE_INFO_TABLE_HINT = _("\n\nHint: [d]efault, [e]nabled, [x]disabled, [i]nstalled, [a]ctive")
 
+
+def _profile_comparison_key(profile):
+    return profile.getName()
+
+
 class ModuleBase(object):
     # :api
 
@@ -68,7 +75,7 @@ class ModuleBase(object):
         # <package_name, set_of_spec>
         fail_safe_repo = hawkey.MODULE_FAIL_SAFE_REPO_NAME
         install_dict = {}
-        install_set_artefacts = set()
+        install_set_artifacts = set()
         fail_safe_repo_used = False
         for spec, (nsvcap, moduledict) in module_dicts.items():
             for name, streamdict in moduledict.items():
@@ -76,6 +83,8 @@ class ModuleBase(object):
                     install_module_list = [x for x in module_list
                                            if self.base._moduleContainer.isModuleActive(x.getId())]
                     if not install_module_list:
+                        logger.error(_("All matches for argument '{0}' in module '{1}:{2}' are not "
+                                       "active").format(spec, name, stream))
                         error_specs.append(spec)
                         continue
                     profiles = []
@@ -90,8 +99,8 @@ class ModuleBase(object):
                         if not profiles:
                             available_profiles = latest_module.getProfiles()
                             if available_profiles:
-                                profile_names = ", ".join(
-                                    [profile.getName() for profile in available_profiles])
+                                profile_names = ", ".join(sorted(
+                                    [profile.getName() for profile in available_profiles]))
                                 msg = _("Unable to match profile for argument {}. Available "
                                         "profiles for '{}:{}': {}").format(
                                     spec, name, stream, profile_names)
@@ -106,22 +115,22 @@ class ModuleBase(object):
                         if not profiles_strings:
                             available_profiles = latest_module.getProfiles()
                             if available_profiles:
-                                profile_names = ", ".join(
-                                    [profile.getName() for profile in available_profiles])
+                                profile_names = ", ".join(sorted(
+                                    [profile.getName() for profile in available_profiles]))
                                 msg = _("No default profiles for module {}:{}. Available profiles"
                                         ": {}").format(
                                     name, stream, profile_names)
                             else:
-                                msg = _("No default profiles for module {}:{}").format(name, stream)
+                                msg = _("No profiles for module {}:{}").format(name, stream)
                             logger.error(msg)
-                            no_match_specs.append(spec)
+                            error_specs.append(spec)
                         for profile in set(profiles_strings):
                             module_profiles = latest_module.getProfiles(profile)
                             if not module_profiles:
                                 logger.error(
                                     _("Default profile {} not available in module {}:{}").format(
                                         profile, name, stream))
-                                no_match_specs.append(spec)
+                                error_specs.append(spec)
 
                             profiles.extend(module_profiles)
                     for profile in profiles:
@@ -129,34 +138,146 @@ class ModuleBase(object):
                         for pkg_name in profile.getContent():
                             install_dict.setdefault(pkg_name, set()).add(spec)
                     for module in install_module_list:
-                        install_set_artefacts.update(module.getArtifacts())
+                        install_set_artifacts.update(module.getArtifacts())
         if fail_safe_repo_used:
             raise dnf.exceptions.Error(_(
                 "Installing module from Fail-Safe repository is not allowed"))
-        install_base_query = self.base.sack.query().filterm(
-            nevra_strict=install_set_artefacts).apply()
+        __, profiles_errors = self._install_profiles_internal(
+            install_set_artifacts, install_dict, strict)
+        if profiles_errors:
+            error_specs.extend(profiles_errors)
 
-        # add hot-fix packages
-        hot_fix_repos = [i.id for i in self.base.repos.iter_enabled() if i.module_hotfixes]
-        hotfix_packages = self.base.sack.query().filterm(reponame=hot_fix_repos).filterm(
-            name=install_dict.keys())
-        install_base_query = install_base_query.union(hotfix_packages)
+        if no_match_specs or error_specs or solver_errors:
+            raise dnf.exceptions.MarkingErrors(no_match_group_specs=no_match_specs,
+                                               error_group_specs=error_specs,
+                                               module_depsolv_errors=solver_errors)
 
-        for pkg_name, set_specs in install_dict.items():
-            query = install_base_query.filter(name=pkg_name)
-            if not query:
-                # package can also be non-modular or part of another stream
-                query = self.base.sack.query().filterm(name=pkg_name)
-                if not query:
-                    for spec in set_specs:
-                        logger.error(_("Unable to resolve argument {}").format(spec))
-                    logger.error(_("No match for package {}").format(pkg_name))
-                    error_specs.extend(set_specs)
-                    continue
-            self.base._goal.group_members.add(pkg_name)
+    def switch_to(self, module_specs, strict=True):
+        # :api
+        no_match_specs, error_specs, module_dicts = self._resolve_specs_enable(module_specs)
+        # collect name of artifacts from new modules for distrosync
+        new_artifacts_names = set()
+        # collect name of artifacts from active modules for distrosync before sack update
+        active_artifacts_names = set()
+        src_arches = {"nosrc", "src"}
+        for spec, (nsvcap, moduledict) in module_dicts.items():
+            for name in moduledict.keys():
+                for module in self.base._moduleContainer.query(name, "", "", "", ""):
+                    if self.base._moduleContainer.isModuleActive(module):
+                        for artifact in module.getArtifacts():
+                            arch = artifact.rsplit(".", 1)[1]
+                            if arch in src_arches:
+                                continue
+                            pkg_name = artifact.rsplit("-", 2)[0]
+                            active_artifacts_names.add(pkg_name)
+
+        solver_errors = self._update_sack()
+
+        dependency_error_spec = self._enable_dependencies(module_dicts)
+        if dependency_error_spec:
+            error_specs.extend(dependency_error_spec)
+
+        # <package_name, set_of_spec>
+        fail_safe_repo = hawkey.MODULE_FAIL_SAFE_REPO_NAME
+        install_dict = {}
+        install_set_artifacts = set()
+        fail_safe_repo_used = False
+
+        # list of name: [profiles] for module profiles being removed
+        removed_profiles = self.base._moduleContainer.getRemovedProfiles()
+
+        for spec, (nsvcap, moduledict) in module_dicts.items():
+            for name, streamdict in moduledict.items():
+                for stream, module_list in streamdict.items():
+                    install_module_list = [x for x in module_list
+                                           if self.base._moduleContainer.isModuleActive(x.getId())]
+                    if not install_module_list:
+                        "No active matches for argument '{0}' in module '{1}:{2}'"
+                        logger.error(_("No active matches for argument '{0}' in module "
+                                       "'{1}:{2}'").format(spec, name, stream))
+                        error_specs.append(spec)
+                        continue
+                    profiles = []
+                    latest_module = self._get_latest(install_module_list)
+                    if latest_module.getRepoID() == fail_safe_repo:
+                        msg = _(
+                            "Installing module '{0}' from Fail-Safe repository {1} is not allowed")
+                        logger.critical(msg.format(latest_module.getNameStream(), fail_safe_repo))
+                        fail_safe_repo_used = True
+                    if nsvcap.profile:
+                        profiles.extend(latest_module.getProfiles(nsvcap.profile))
+                        if not profiles:
+                            available_profiles = latest_module.getProfiles()
+                            if available_profiles:
+                                profile_names = ", ".join(sorted(
+                                    [profile.getName() for profile in available_profiles]))
+                                msg = _("Unable to match profile for argument {}. Available "
+                                        "profiles for '{}:{}': {}").format(
+                                    spec, name, stream, profile_names)
+                            else:
+                                msg = _("Unable to match profile for argument {}").format(spec)
+                            logger.error(msg)
+                            no_match_specs.append(spec)
+                            continue
+                    elif name in removed_profiles:
+
+                        for profile in removed_profiles[name]:
+                            module_profiles = latest_module.getProfiles(profile)
+                            if not module_profiles:
+                                logger.warning(
+                                    _("Installed profile '{0}' is not available in module "
+                                      "'{1}' stream '{2}'").format(profile, name, stream))
+                                continue
+                            profiles.extend(module_profiles)
+                    for profile in profiles:
+                        self.base._moduleContainer.install(latest_module, profile.getName())
+                        for pkg_name in profile.getContent():
+                            install_dict.setdefault(pkg_name, set()).add(spec)
+                    for module in install_module_list:
+                        artifacts = module.getArtifacts()
+                        install_set_artifacts.update(artifacts)
+                        for artifact in artifacts:
+                            arch = artifact.rsplit(".", 1)[1]
+                            if arch in src_arches:
+                                continue
+                            pkg_name = artifact.rsplit("-", 2)[0]
+                            new_artifacts_names.add(pkg_name)
+        if fail_safe_repo_used:
+            raise dnf.exceptions.Error(_(
+                "Installing module from Fail-Safe repository is not allowed"))
+        install_base_query, profiles_errors = self._install_profiles_internal(
+            install_set_artifacts, install_dict, strict)
+        if profiles_errors:
+            error_specs.extend(profiles_errors)
+
+        # distrosync module name
+        all_names = set()
+        all_names.update(new_artifacts_names)
+        all_names.update(active_artifacts_names)
+        remove_query = self.base.sack.query().filterm(empty=True)
+        base_no_source_query = self.base.sack.query().filterm(arch__neq=['src', 'nosrc']).apply()
+
+        for pkg_name in all_names:
+            query = base_no_source_query.filter(name=pkg_name)
+            installed = query.installed()
+            if not installed:
+                continue
+            available = query.available()
+            if not available:
+                logger.warning(_("No packages available to distrosync for package name "
+                                 "'{}'").format(pkg_name))
+                if pkg_name not in new_artifacts_names:
+                    remove_query = remove_query.union(query)
+                continue
+
+            only_new_module = query.intersection(install_base_query)
+            if only_new_module:
+                query = only_new_module
             sltr = dnf.selector.Selector(self.base.sack)
             sltr.set(pkg=query)
-            self.base._goal.install(select=sltr, optional=(not strict))
+            self.base._goal.distupgrade(select=sltr)
+        self.base._remove_if_unneeded(remove_query)
+
         if no_match_specs or error_specs or solver_errors:
             raise dnf.exceptions.MarkingErrors(no_match_group_specs=no_match_specs,
                                                error_group_specs=error_specs,
@@ -175,6 +296,9 @@ class ModuleBase(object):
         fail_safe_repo = hawkey.MODULE_FAIL_SAFE_REPO_NAME
         fail_safe_repo_used = False
 
+        #  Remove source packages because they cannot be installed or upgraded
+        base_no_source_query = self.base.sack.query().filterm(arch__neq=['src', 'nosrc']).apply()
+
         for spec in module_specs:
             module_list, nsvcap = self._get_modules(spec)
             if not module_list:
@@ -185,7 +309,7 @@ class ModuleBase(object):
             if not update_module_list:
                 logger.error(_("Unable to resolve argument {}").format(spec))
                 continue
-            module_dict = self._create_module_dict_and_enable(update_module_list, False)
+            module_dict = self._create_module_dict_and_enable(update_module_list, spec, False)
             upgrade_package_set = set()
             for name, streamdict in module_dict.items():
                 for stream, module_list_from_dict in streamdict.items():
@@ -206,15 +330,15 @@ class ModuleBase(object):
                     else:
                         for profile in latest_module.getProfiles():
                             upgrade_package_set.update(profile.getContent())
-                        for artefact in latest_module.getArtifacts():
-                            subj = hawkey.Subject(artefact)
+                        for artifact in latest_module.getArtifacts():
+                            subj = hawkey.Subject(artifact)
                             for nevra_obj in subj.get_nevra_possibilities(
                                     forms=[hawkey.FORM_NEVRA]):
                                 upgrade_package_set.add(nevra_obj.name)
 
             if not upgrade_package_set:
                 logger.error(_("Unable to match profile in argument {}").format(spec))
-            query = self.base.sack.query().filterm(name=upgrade_package_set)
+            query = base_no_source_query.filter(name=upgrade_package_set)
             if query:
                 sltr = dnf.selector.Selector(self.base.sack)
                 sltr.set(pkg=query)
@@ -234,7 +358,7 @@ class ModuleBase(object):
             if not module_list:
                 no_match_specs.append(spec)
                 continue
-            module_dict = self._create_module_dict_and_enable(module_list, False)
+            module_dict = self._create_module_dict_and_enable(module_list, spec, False)
             remove_packages_names = []
             for name, streamdict in module_dict.items():
                 for stream, module_list_from_dict in streamdict.items():
@@ -278,11 +402,11 @@ class ModuleBase(object):
         if module_list:
             latest = module_list[0]
             for module in module_list[1:]:
-                if module.getVersion() > latest.getVersion():
+                if module.getVersionNum() > latest.getVersionNum():
                     latest = module
         return latest
 
-    def _create_module_dict_and_enable(self, module_list, enable=True):
+    def _create_module_dict_and_enable(self, module_list, spec, enable=True):
         moduleDict = {}
         for module in module_list:
             moduleDict.setdefault(
@@ -293,7 +417,14 @@ class ModuleBase(object):
             if len(streamDict) > 1:
                 if moduleState != STATE_DEFAULT and moduleState != STATE_ENABLED \
                         and moduleState != STATE_DISABLED:
-                    raise EnableMultipleStreamsException(moduleName)
+                    streams_str = "', '".join(
+                        sorted(streamDict.keys(), key=functools.cmp_to_key(self.base.sack.evr_cmp)))
+                    msg = _("Argument '{argument}' matches {stream_count} streams ('{streams}') of "
+                            "module '{module}', but non of the streams are enabled or "
+                            "default").format(
+                        argument=spec, stream_count=len(streamDict), streams=streams_str,
+                        module=moduleName)
+                    raise EnableMultipleStreamsException(moduleName, msg)
                 if moduleState == STATE_ENABLED:
                     stream = self.base._moduleContainer.getEnabledStream(moduleName)
                 else:
@@ -312,7 +443,7 @@ class ModuleBase(object):
             assert len(streamDict) == 1
         return moduleDict
 
-    def _resolve_specs_enable_update_sack(self, module_specs):
+    def _resolve_specs_enable(self, module_specs):
         no_match_specs = []
         error_spec = []
         module_dicts = {}
@@ -322,20 +453,27 @@ class ModuleBase(object):
                 no_match_specs.append(spec)
                 continue
             try:
-                module_dict = self._create_module_dict_and_enable(module_list, True)
+                module_dict = self._create_module_dict_and_enable(module_list, spec, True)
                 module_dicts[spec] = (nsvcap, module_dict)
             except (RuntimeError, EnableMultipleStreamsException) as e:
                 error_spec.append(spec)
                 logger.error(ucd(e))
                 logger.error(_("Unable to resolve argument {}").format(spec))
+        return no_match_specs, error_spec, module_dicts
+
+    def _update_sack(self):
         hot_fix_repos = [i.id for i in self.base.repos.iter_enabled() if i.module_hotfixes]
         try:
             solver_errors = self.base.sack.filter_modules(
                 self.base._moduleContainer, hot_fix_repos, self.base.conf.installroot,
-                self.base.conf.module_platform_id,
-                self.base.conf.debug_solver)
+                self.base.conf.module_platform_id, update_only=True,
+                debugsolver=self.base.conf.debug_solver)
         except hawkey.Exception as e:
             raise dnf.exceptions.Error(ucd(e))
+        return solver_errors
+
+    def _enable_dependencies(self, module_dicts):
+        error_spec = []
         for spec, (nsvcap, moduleDict) in module_dicts.items():
             for streamDict in moduleDict.values():
                 for modules in streamDict.values():
@@ -346,6 +484,17 @@ class ModuleBase(object):
                         error_spec.append(spec)
                         logger.error(ucd(e))
                         logger.error(_("Unable to resolve argument {}").format(spec))
+        return error_spec
+
+    def _resolve_specs_enable_update_sack(self, module_specs):
+        no_match_specs, error_spec, module_dicts = self._resolve_specs_enable(module_specs)
+
+        solver_errors = self._update_sack()
+
+        dependency_error_spec = self._enable_dependencies(module_dicts)
+        if dependency_error_spec:
+            error_spec.extend(dependency_error_spec)
+
         return no_match_specs, error_spec, solver_errors, module_dicts
 
     def _modules_reset_or_disable(self, module_specs, to_state):
@@ -368,14 +517,7 @@ class ModuleBase(object):
                 if to_state == STATE_DISABLED:
                     self.base._moduleContainer.disable(name)
 
-        hot_fix_repos = [i.id for i in self.base.repos.iter_enabled() if i.module_hotfixes]
-        try:
-            solver_errors = self.base.sack.filter_modules(
-                self.base._moduleContainer, hot_fix_repos, self.base.conf.installroot,
-                self.base.conf.module_platform_id, update_only=True,
-                debugsolver=self.base.conf.debug_solver)
-        except hawkey.Exception as e:
-            raise dnf.exceptions.Error(ucd(e))
+        solver_errors = self._update_sack()
         return no_match_specs, solver_errors
 
     def _get_package_name_set_and_remove_profiles(self, module_list, nsvcap, remove=False):
@@ -418,26 +560,26 @@ class ModuleBase(object):
                 lines = OrderedDict()
                 lines["Name"] = module.getFullIdentifier()
 
-                for profile in module.getProfiles():
+                for profile in sorted(module.getProfiles(), key=_profile_comparison_key):
                     lines[profile.getName()] = "\n".join(
                         [pkgName for pkgName in profile.getContent()])
 
                 output.add(self._create_simple_table(lines).toString())
         return "\n\n".join(sorted(output))
 
-    def _profile_report_formater(self, modulePackage, default_profiles, enabled_str):
+    def _profile_report_formatter(self, modulePackage, default_profiles, enabled_str):
         installed_profiles = self.base._moduleContainer.getInstalledProfiles(
             modulePackage.getName())
         available_profiles = modulePackage.getProfiles()
         profiles_str = ""
-        for profile in available_profiles:
+        for profile in sorted(available_profiles, key=_profile_comparison_key):
             profiles_str += "{}{}".format(
                 profile.getName(), " [d]" if profile.getName() in default_profiles else "")
             profiles_str += " [i], " if profile.getName() in installed_profiles and enabled_str \
                 else ", "
         return profiles_str[:-2]
 
-    def _module_strs_formater(self, modulePackage, markActive=False):
+    def _module_strs_formatter(self, modulePackage, markActive=False):
         default_str = ""
         enabled_str = ""
         disabled_str = ""
@@ -470,12 +612,12 @@ class ModuleBase(object):
                 logger.info(_("Ignoring unnecessary profile: '{}/{}'").format(
                     nsvcap.name, nsvcap.profile))
             for modulePackage in module_list:
-                default_str, enabled_str, disabled_str = self._module_strs_formater(
+                default_str, enabled_str, disabled_str = self._module_strs_formatter(
                     modulePackage, markActive=True)
                 default_profiles = self.base._moduleContainer.getDefaultProfiles(
                     modulePackage.getName(), modulePackage.getStream())
 
-                profiles_str = self._profile_report_formater(
+                profiles_str = self._profile_report_formatter(
                     modulePackage, default_profiles, enabled_str)
 
                 lines = OrderedDict()
@@ -490,6 +632,12 @@ class ModuleBase(object):
                 lines["Repo"] = modulePackage.getRepoID()
                 lines["Summary"] = modulePackage.getSummary()
                 lines["Description"] = modulePackage.getDescription()
+                req_set = set()
+                for req in modulePackage.getModuleDependencies():
+                    for require_dict in req.getRequires():
+                        for mod_require, stream in require_dict.items():
+                            req_set.add("{}:[{}]".format(mod_require, ",".join(stream)))
+                lines["Requires"] = "\n".join(sorted(req_set))
                 lines["Artifacts"] = "\n".join(sorted(modulePackage.getArtifacts()))
                 output.add(self._create_simple_table(lines).toString())
         str_table = "\n\n".join(sorted(output))
@@ -510,10 +658,11 @@ class ModuleBase(object):
         column_value.setNewlineWrapFunction()
 
         for line_name, value in lines.items():
-            if value:
-                line = table.newLine()
-                line.getColumnCell(column_name).setData(line_name)
-                line.getColumnCell(column_value).setData(str(value))
+            if value is None:
+                value = ""
+            line = table.newLine()
+            line.getColumnCell(column_name).setData(line_name)
+            line.getColumnCell(column_value).setData(str(value))
 
         return table
 
@@ -539,10 +688,14 @@ class ModuleBase(object):
         output = set()
         modulePackages = self.base._moduleContainer.getModulePackages()
         baseQuery = self.base.sack.query().filterm(empty=True).apply()
+        getBestInitQuery = self.base.sack.query(flags=hawkey.IGNORE_MODULAR_EXCLUDES)
+
         for spec in rpm_specs:
             subj = dnf.subject.Subject(spec)
             baseQuery = baseQuery.union(subj.get_best_query(
-                self.base.sack, with_nevra=True, with_provides=False, with_filenames=False))
+                self.base.sack, with_nevra=True, with_provides=False, with_filenames=False,
+                query=getBestInitQuery))
+
         baseQuery.apply()
 
         for modulePackage in modulePackages:
@@ -554,12 +707,12 @@ class ModuleBase(object):
                 for pkg in query:
                     string_output = ""
                     profiles = []
-                    for profile in modulePackage.getProfiles():
+                    for profile in sorted(modulePackage.getProfiles(), key=_profile_comparison_key):
                         if pkg.name in profile.getContent():
                             profiles.append(profile.getName())
                     lines = OrderedDict()
                     lines["Module"] = modulePackage.getFullIdentifier()
-                    lines["Profiles"] = " ".join(profiles)
+                    lines["Profiles"] = " ".join(sorted(profiles))
                     lines["Repo"] = modulePackage.getRepoID()
                     lines["Summary"] = modulePackage.getSummary()
 
@@ -597,11 +750,11 @@ class ModuleBase(object):
                     else:
                         modulePackage = nameStreamArch[0]
                 line = table.newLine()
-                default_str, enabled_str, disabled_str = self._module_strs_formater(
+                default_str, enabled_str, disabled_str = self._module_strs_formatter(
                     modulePackage, markActive=False)
                 default_profiles = self.base._moduleContainer.getDefaultProfiles(
                     modulePackage.getName(), modulePackage.getStream())
-                profiles_str = self._profile_report_formater(modulePackage, default_profiles,
+                profiles_str = self._profile_report_formatter(modulePackage, default_profiles,
                                                              enabled_str)
                 line.getColumnCell(column_name).setData(modulePackage.getName())
                 line.getColumnCell(
@@ -661,6 +814,35 @@ class ModuleBase(object):
     def _format_repoid(self, repo_name):
         return "{}\n".format(self.base.output.term.bold(repo_name))
 
+    def _install_profiles_internal(self, install_set_artifacts, install_dict, strict):
+        #  Remove source packages because they cannot be installed or upgraded
+        base_no_source_query = self.base.sack.query().filterm(arch__neq=['src', 'nosrc']).apply()
+        install_base_query = base_no_source_query.filter(nevra_strict=install_set_artifacts)
+        error_specs = []
+
+        # add hot-fix packages
+        hot_fix_repos = [i.id for i in self.base.repos.iter_enabled() if i.module_hotfixes]
+        hotfix_packages = base_no_source_query.filter(
+            reponame=hot_fix_repos, name=install_dict.keys())
+        install_base_query = install_base_query.union(hotfix_packages)
+
+        for pkg_name, set_specs in install_dict.items():
+            query = install_base_query.filter(name=pkg_name)
+            if not query:
+                # package can also be non-modular or part of another stream
+                query = base_no_source_query.filter(name=pkg_name)
+                if not query:
+                    for spec in set_specs:
+                        logger.error(_("Unable to resolve argument {}").format(spec))
+                    logger.error(_("No match for package {}").format(pkg_name))
+                    error_specs.extend(set_specs)
+                    continue
+            self.base._goal.group_members.add(pkg_name)
+            sltr = dnf.selector.Selector(self.base.sack)
+            sltr.set(pkg=query)
+            self.base._goal.install(select=sltr, optional=(not strict))
+        return install_base_query, error_specs
+
 
 def format_modular_solver_errors(errors):
     msg = dnf.util._format_resolve_problems(errors)
